from dataclasses import dataclass, field
from datetime import datetime, timedelta
from typing import Any, Dict, List, Union

from .abi import (
    AbiBytes, BaseAbiObject, UInt8, UInt16, UInt32, VarUInt, Name, TimePointSec
)
from .authority import EosPermissionLevel


AbiActionPayload = Union[Dict[str, Any], AbiBytes, BaseAbiObject]


@dataclass
class EosAction(BaseAbiObject):
    account: Name
    name: Name
    authorization: List[EosPermissionLevel]
    data: AbiActionPayload


@dataclass
class EosExtension(BaseAbiObject):
    extension_type: UInt16
    data: AbiBytes


@dataclass
class EosTransaction(BaseAbiObject):
    # TAPOS fields
    expiration: TimePointSec = field(
        default_factory=lambda: datetime.now() + timedelta(seconds=120)
    )
    ref_block_num: UInt16 = 0
    ref_block_prefix: UInt32 = 0

    max_net_usage_words: VarUInt = 0
    max_cpu_usage_ms: UInt8 = 0
    delay_sec: VarUInt = 0
    context_free_actions: List[EosAction] = field(default_factory=list)
    actions: List[EosAction] = field(default_factory=list)
    transaction_extensions: List[EosExtension] = field(default_factory=list)



from dataclasses import dataclass, field
from typing import List

from .abi import AbiBytes, BaseAbiObject, UInt16, UInt32, Name


@dataclass
class EosPermissionLevel(BaseAbiObject):
    actor: Name
    permission: Name


@dataclass
class EosKeyWeight(BaseAbiObject):
    key: AbiBytes
    weight: UInt16


@dataclass
class EosPermissionLevelWeight(BaseAbiObject):
    permission: EosPermissionLevel
    weight: UInt16


@dataclass
class EosWaitWeight(BaseAbiObject):
    wait_sec: UInt32
    weight: UInt16


@dataclass
class EosAuthority(BaseAbiObject):
    threshold: UInt32 = 1
    keys: List[EosKeyWeight] = field(default_factory=list)
    accounts: List[EosPermissionLevelWeight] = field(default_factory=list)
    waits: List[EosWaitWeight] = field(default_factory=list)



from typing import Optional

from aioeosabi.keys import EosKey
from aioeosabi.types import EosPermissionLevel, EosPermissionLevelWeight


class EosAccount:
    """
    Describes account on EOS blockchain. Contrary to other blockchains such as
    Bitcoin or Ethereum, public key is not an address. An account can have
    multiple keys with certain permissions, the default ones are "active" and
    "owner". These keys are specified in the transaction which creates the
    account in the first place.

    Class provides a set of helper methods for generating objects such as
    authorizations.

    :param name: name of the account,

    Please provide key in one of the formats - EosKey instance, private key in
    WIF or PVT format and public key. Only one of these is accepted. If no key
    is provided, a new one will be generated.

    ;param key: EosKey instance,
    :param private_key: private key in WIF or PVT format,
    :param public_key: public key
    """

    def __init__(
        self,
        name: str,
        *,
        key: Optional[EosKey] = None,
        private_key: str = '',
        public_key: str = '',
    ):
        assert name, "Account name can't be empty"

        # While this could be written as some kind of binary operation and it
        # would be clever and cool, this is more clear
        assert any((
            key and not private_key and not public_key,
            not key and private_key and not public_key,
            not key and not private_key and public_key,
            not key and not private_key and not public_key
        )), 'Exactly 1 key is required'

        self.name = name
        self.key = (
            key or EosKey(public_key=public_key, private_key=private_key)
        )

    def authorization(self, permission: str) -> EosPermissionLevel:
        """Creates authorization object with given permission"""
        return EosPermissionLevel(actor=self.name, permission=permission)

    def permission_level_weight(
        self, permission: str, weight: int
    ) -> EosPermissionLevelWeight:
        return EosPermissionLevelWeight(
            permission=self.authorization(permission),
            weight=weight
        )



from . import serializer, exceptions  # noqa
from .account import EosAccount  # noqa
from .keys import EosKey  # noqa
from .rpc import EosJsonRpc  # noqa
from .types import (
    UInt8, UInt16, UInt32, UInt64, Int8, Int16, Int32, Int64, VarUInt, Float32,
    Float64, TimePointSec, TimePoint, Name, AbiBytes, BaseAbiObject,
    is_abi_object, EosPermissionLevel, EosKeyWeight, EosPermissionLevelWeight,
    EosWaitWeight, EosAuthority, AbiActionPayload, EosAction, EosTransaction
)  # noqa

__all__ = [
    'serializer',
    'exceptions',

    # Account
    'EosAccount',

    # Keys
    'EosKey',

    # RPC
    'EosJsonRpc',

    # types
    # base ABI types
    'UInt8', 'UInt16', 'UInt32', 'UInt64', 'Int8', 'Int16', 'Int32', 'Int64',
    'VarUInt', 'Float32', 'Float64', 'TimePointSec', 'TimePoint', 'Name',
    'AbiBytes', 'BaseAbiObject', 'is_abi_object',

    # authority
    'EosPermissionLevel', 'EosKeyWeight', 'EosPermissionLevelWeight',
    'EosWaitWeight', 'EosAuthority',

    # transaction
    'AbiActionPayload', 'EosAction', 'EosTransaction'
]



from dataclasses import dataclass, fields
from datetime import datetime
import inspect
from typing import Any, ClassVar, Dict, NewType, TYPE_CHECKING


# EOS ABI types
# TODO: implement bool, uint128, int128, float128, block_timestamp_type,
# symbol, symbol_code, asset, checksum160, checksum256, checksum512,
# public_key, private_key, signature, extended_asset
if TYPE_CHECKING:
    UInt8 = int
    UInt16 = int
    UInt32 = int
    UInt64 = int
    Int8 = int
    Int16 = int
    Int32 = int
    Int64 = int
    VarUInt = int
    Float32 = float
    Float64 = float
    TimePointSec = datetime
    TimePoint = datetime
    Name = str
    AbiBytes = bytes
else:
    # Our runtime logic depends on these being new types, but this makes mypy
    # require explicit casting
    UInt8 = NewType('UInt8', int)
    UInt16 = NewType('UInt16', int)
    UInt32 = NewType('UInt32', int)
    UInt64 = NewType('UInt64', int)
    Int8 = NewType('Int8', int)
    Int16 = NewType('Int16', int)
    Int32 = NewType('Int32', int)
    Int64 = NewType('Int64', int)
    Float32 = NewType('Float32', float)
    Float64 = NewType('Float64', float)
    Name = NewType('Name', str)
    TimePointSec = NewType('TimePointSec', datetime)
    TimePoint = NewType('TimePoint', datetime)
    AbiBytes = NewType('AbiBytes', bytes)

    # this type is weird because it's like int, but it has no fixed size
    VarUInt = NewType('VarUInt', int)


@dataclass
class BaseAbiObject:
    __dataclass_fields__: ClassVar[Dict]

    @classmethod
    def _serializable_fields(cls):
        return (x.name for x in fields(cls))


def is_abi_object(obj: Any) -> bool:
    """Object is an ABI object if it's a subclass of BaseAbiObject"""
    is_class = inspect.isclass(obj)
    return is_class and issubclass(obj, BaseAbiObject)



class EosRpcException(Exception):
    """Base EOS exception"""


class EosAccountExistsException(EosRpcException):
    """Thrown by create_wallet where account with given name already exists"""


class EosAccountDoesntExistException(EosRpcException):
    """Thrown by get_account where account doesn't exist"""


class EosActionValidateException(EosRpcException):
    """Raised when action payload is invalid"""


class EosMissingTaposFieldsException(EosRpcException):
    """TAPOS fields are missing from Transaction object"""


class EosDeadlineException(EosRpcException):
    """Transaction timed out"""


class EosTxCpuUsageExceededException(EosRpcException):
    """Not enough EOS were staked for CPU"""


class EosTxNetUsageExceededException(EosRpcException):
    """Not enough EOS were staked for NET"""


class EosRamUsageExceededException(EosRpcException):
    """Transaction requires more RAM than what's available on the account"""


class EosAssertMessageException(EosRpcException):
    """
    Generic assertion error from smart contract, can mean literally anything,
    need to parse C++ traceback to figure out what went wrong.
    """


class EosSerializerException(Exception):
    """Base exception class for serializer errors"""


class EosSerializerUnsupportedTypeException(EosSerializerException):
    """Our serializer doesn't support provided object type"""


class EosSerializerAbiNameTooLongException(EosSerializerException):
    def __init__(self):
        super().__init__('Value is too long, expected up to 13 characters')


class EosSerializerAbiNameInvalidCharactersException(EosSerializerException):
    def __init__(self):
        super().__init__('Value contains invalid characters')



"""Helpers for creating actions on eosio.token contract"""
from aioeosabi import types


def transfer(
    from_addr: str,
    to_addr: str,
    quantity: str,
    memo: str = '',
    authorization=[]
) -> types.EosAction:
    return types.EosAction(
        account='eosio.token',
        name='transfer',
        authorization=authorization,
        data={
            'from': from_addr,
            'to': to_addr,
            'quantity': quantity,
            'memo': memo
        }
    )


def close(owner, symbol, authorization=[]) -> types.EosAction:
    return types.EosAction(
        account='eosio.token',
        name='close',
        authorization=authorization,
        data={
            'owner': owner,
            'symbol': symbol
        }
    )



import re
import hashlib
import secrets

import base58
import ecdsa

from aioeosabi.types import EosKeyWeight
from Crypto.Hash import RIPEMD

class EosKey:
    """
    EosKey instance.

    Depends on which kwargs are given, this works in a different way:
    - No kwargs - generates a new private key
    - Only private_key - public key is being derived from private key
    - Only public_key - EosKey instance has no private key
    """

    def __init__(self, *, private_key: str = None, public_key: str = None):
        assert not (private_key and public_key), 'Pass only 1 key'
        if private_key:
            private_key = self._parse_key(private_key)
            self._sk = ecdsa.SigningKey.from_string(
                private_key, curve=ecdsa.SECP256k1
            )
        elif not public_key:
            entropy = ecdsa.util.PRNG(secrets.randbits(512))
            self._sk = ecdsa.SigningKey.generate(
                entropy=entropy, curve=ecdsa.SECP256k1
            )
        self._vk = public_key if public_key else self._sk.get_verifying_key()

    def _parse_key(self, private_str):
        """
        EOS has two private key formats.
        - WIF - legacy format, has to start with 0x80 byte,
        - PVT - PVT_{key type}_{base58 encoded key with checksum}
        """
        # check if key is in PVT format
        match = re.search('^PVT_([A-Za-z0-9]+)_([A-Za-z0-9]+)$', private_str)
        if match:
            key_type, key_string = match.groups()
            return self._check_decode(key_string, key_type)

        # fallback to WIF
        private_key = self._check_decode(private_str, 'sha256x2')
        if private_key[0] != 0x80:
            raise ValueError('Invalid version')
        return private_key[1:]

    def _calculate_checksum(self, key, key_type=''):
        """
        Takes a private key, returns a checksum.

        ``key_type`` determines the kind of checksum:
        - sha256x2
        - K1 - rmd160 checksum
        - empty string - rmd160 checksum without suffix
        """
        if key_type == 'sha256x2':
            first_sha = hashlib.sha256(key).digest()
            return hashlib.sha256(first_sha).digest()[:4]
        elif key_type in ('', 'K1'):
            r = RIPEMD.new()
            r.update(key + key_type.encode('utf-8'))
            return r.digest()[:4]
        else:
            raise TypeError('Unsupported key type {}'.format(key_type))

    def _check_encode(self, key_buffer, key_type=''):
        """
        Encodes the key to checksummed base58 format. ``key_type`` determines
        checksum type.
        """
        # base58.b58encode only takes bytes, not bytearray, make sure we have
        # the right type
        assert type(key_buffer) in (bytes, bytearray)
        if isinstance(key_buffer, bytearray):
            key_buffer = bytes(key_buffer)

        checksum = self._calculate_checksum(key_buffer, key_type)

        # b58encode returns bytes, we always cast this to regular strings in
        # a next call so let's do this here
        return base58.b58encode(key_buffer + checksum).decode()

    def _check_decode(self, key_string, key_type=''):
        """
        Decodes the key from checksummed base58 format, checks it against
        expected checksum and returns the value. ``key_type`` determines
        checksum type.
        """
        buffer = base58.b58decode(key_string)
        key, checksum = buffer[:-4], buffer[-4:]
        new_checksum = self._calculate_checksum(key, key_type)

        if checksum != new_checksum:
            raise ValueError(
                f'checksums do not match: {checksum} != {new_checksum}'
            )
        return key

    def _recover_key(self, digest, signature, i):
        ''' Recover the public key from the sig
            http://www.secg.org/sec1-v2.pdf
        '''
        curve = ecdsa.SECP256k1.curve
        G = ecdsa.SECP256k1.generator
        order = ecdsa.SECP256k1.order
        yp = (i % 2)
        r, s = ecdsa.util.sigdecode_string(signature, order)
        x = r + (i // 2) * order
        alpha = ((x * x * x) + (curve.a() * x) + curve.b()) % curve.p()
        beta = ecdsa.numbertheory.square_root_mod_prime(alpha, curve.p())
        y = beta if (beta - yp) % 2 == 0 else curve.p() - beta
        # generate R
        R = ecdsa.ellipticcurve.Point(curve, x, y, order)
        e = ecdsa.util.string_to_number(digest)
        # compute Q
        Q = (
            ecdsa.numbertheory.inverse_mod(r, order)
            * (s * R + (-e % order) * G)
        )
        # verify message
        verifying_key = ecdsa.VerifyingKey.from_public_point(
            Q, curve=ecdsa.SECP256k1
        )
        if not verifying_key.verify_digest(
            signature, digest, sigdecode=ecdsa.util.sigdecode_string
        ):
            return None
        return ecdsa.VerifyingKey.from_public_point(Q, curve=ecdsa.SECP256k1)

    def _recovery_pubkey_param(self, digest, signature):
        """
        Use to derive a number that will allow for the easy recovery of the
        public key from the signature
        """
        for i in range(0, 4):
            p = self._recover_key(digest, signature, i)
            if p.to_string() == self._vk.to_string():
                return i

    def to_public(self):
        """Returns compressed, base58 encoded public key prefixed with EOS"""
        compressed = self._vk.to_string(encoding='compressed')
        return f'EOS{self._check_encode(compressed)}'

    def to_wif(self):
        """Converts private key to legacy WIF format"""
        private_key = b'\x80' + self._sk.to_string()
        return self._check_encode(private_key, 'sha256x2')

    def to_pvt(self, key_type='K1'):
        """Converts private key to PVT format"""
        private_key = self._check_encode(self._sk.to_string(), key_type)
        return f'PVT_{key_type}_{private_key}'

    def sign(self, digest):
        """
        Signs sha256 hash with private key. Returns signature in format:
        ``SIG_K1_{digest}``
        """
        cnt = 0
        if len(digest) != 32:
            raise ValueError("32 byte buffer required")

        # repeat until the signature is canonical
        is_canonical = False
        while not is_canonical:
            # get deterministic k
            sha_digest = hashlib.sha256(
                digest + bytes([cnt] if cnt else [])
            ).digest()

            k = ecdsa.rfc6979.generate_k(
                self._sk.curve.generator.order(),
                self._sk.privkey.secret_multiplier,
                hashlib.sha256,
                sha_digest
            )
            # sign the message, returns der_signature
            der_signature = self._sk.sign_digest(
                digest, sigencode=ecdsa.util.sigencode_der, k=k
            )

            # decode der_signature
            r, s = ecdsa.util.sigdecode_der(
                der_signature, self._sk.curve.generator.order()
            )
            is_canonical = (
                der_signature[3] == 32
                and der_signature[5 + der_signature[3]] == 32
            )
            # try another one if not canonical
            cnt += 1

        # encode signature in string format and derive recovery parameter
        sig = ecdsa.util.sigencode_string(
            r, s, self._sk.curve.generator.order()
        )

        # 4 because it's compressed and 27 because it's compact (?)
        # https://github.com/EOSIO/eosjs-ecc/blob/master/src/signature.js#L216
        i = self._recovery_pubkey_param(digest, sig) + 4 + 27
        return f'SIG_K1_{self._check_encode(bytes([i]) + sig, "K1")}'

    def verify(self, encoded_sig, digest) -> bool:
        """Verifies signature with private key"""
        _, key_type, signature = encoded_sig.split('_')
        try:
            sig = self._check_decode(signature, key_type)[1:]
            self._vk.verify_digest(
                sig, digest, sigdecode=ecdsa.util.sigdecode_string
            )
        except (ecdsa.keys.BadSignatureError, TypeError):
            return False
        return True

    def to_key_weight(self, weight: int) -> EosKeyWeight:
        return EosKeyWeight(key=self.to_public(), weight=weight)

    def __eq__(self, other) -> bool:
        assert isinstance(other, EosKey), 'Can compare only to EosKey instance'
        return self.to_public() == other.to_public()



"""Helpers for creating actions on eosio contract"""
from typing import Optional

from aioeosabi.types import EosAction, EosAuthority


def newaccount(
    creator,
    account_name,
    *,
    owner: EosAuthority,
    active: Optional[EosAuthority] = None,
    authorization=[]
) -> EosAction:
    return EosAction(
        account='eosio',
        name='newaccount',
        authorization=authorization,
        data={
            'creator': creator,
            'name': account_name,
            'owner': owner,
            'active': active if active else owner
        }
    )


def buyrambytes(payer, receiver, amount, authorization=[]) -> EosAction:
    return EosAction(
        account='eosio',
        name='buyrambytes',
        authorization=authorization,
        data={
            'payer': payer,
            'receiver': receiver,
            'bytes': amount
        }
    )


def sellram(account, amount, authorization=[]) -> EosAction:
    return EosAction(
        account='eosio',
        name='sellram',
        authorization=authorization,
        data={
            'account': account,
            'bytes': amount
        }
    )


def delegatebw(
    from_account, receiver, stake_net_quantity, stake_cpu_quantity,
    transfer=False, authorization=[]
) -> EosAction:
    return EosAction(
        account='eosio',
        name='delegatebw',
        authorization=authorization,
        data={
            'from': from_account,
            'receiver': receiver,
            'stake_net_quantity': stake_net_quantity,
            'stake_cpu_quantity': stake_cpu_quantity,
            'transfer': transfer
        }
    )


def undelegatebw(
    from_account, receiver, unstake_net_quantity, unstake_cpu_quantity,
    authorization=[]
) -> EosAction:
    return EosAction(
        account='eosio',
        name='undelegatebw',
        authorization=authorization,
        data={
            'from': from_account,
            'receiver': receiver,
            'unstake_net_quantity': unstake_net_quantity,
            'unstake_cpu_quantity': unstake_cpu_quantity,
        }
    )



import base64
import binascii
from dataclasses import asdict
import hashlib
from typing import Any, List

from aiohttp import ClientSession
from aioeosabi import exceptions, serializer, EosAccount
from aioeosabi.keys import EosKey
from aioeosabi.types import EosTransaction, is_abi_object, EosAction
from antelopy import AbiCache
from antelopy.types.abi import Abi
import secrets
import asyncio

abicache = AbiCache(chain_endpoint="")

ERROR_NAME_MAP = {
    'exception': exceptions.EosRpcException,
    'deadline_exception': exceptions.EosDeadlineException,
    'action_validate_exception': exceptions.EosActionValidateException,
    'tx_cpu_usage_exceeded': exceptions.EosTxCpuUsageExceededException,
    'tx_net_usage_exceeded': exceptions.EosTxNetUsageExceededException,
    'ram_usage_exceeded': exceptions.EosRamUsageExceededException,
    'eosio_assert_message_exception': exceptions.EosAssertMessageException,
}


def mixed_to_dict(payload: Any):
    """
    Recursively converts payload with mixed ABI objects and dicts to dict
    """
    if isinstance(payload, dict):
        return {k: mixed_to_dict(v) for k, v in payload.items()}
    if is_abi_object(type(payload)):
        return asdict(payload)
    return payload


class EosJsonRpc:
    def __init__(self, url):
        self.URL = url
        self._chain_id = None

    async def post(self, endpoint, json={}):
        async with ClientSession() as session:
            async with session.post(
                f'{self.URL}/v1{endpoint}',
                json=json
            ) as res:
                resp_dict = await res.json(content_type=None)

                # Who needs HTTP status codes, am I right? :D
                if resp_dict.get('code') == 500:
                    error = resp_dict.get('error', {})
                    raise ERROR_NAME_MAP.get(
                        error.get('name'),
                        exceptions.EosRpcException
                    )(error)
                return resp_dict

    async def abi_json_to_bin(self, action, use_stored=True):
        if not use_stored or action.account not in abicache._abi_cache:
            # Fetch the ABI first
            await self.async_add_raw_abi(action.account)
        return {"binargs": abicache.serialize_data(action.account, action.name, action.data)}

    async def get_abi(self, account_name: str):
        return await self.post(
            '/chain/get_abi', {
                'account_name': account_name
            }
        )

    async def async_add_raw_abi(self, account_name: str):
        r = await self.post(
            '/chain/get_abi', {
                'account_name': account_name
            }
        )
        if raw_abi := r.get("abi"):
            abicache._abi_cache[account_name] = Abi(name=account_name, **raw_abi)
            return 
        raise ABINotFoundError(f"Couldn't retrieve ABI for {account_name}")




    async def get_account(self, account_name: str):
        return await self.post(
            '/chain/get_account', {
                'account_name': account_name
            }
        )

    async def get_block_header_state(self, block_num_or_id):
        return await self.post(
            '/chain/get_block_header_state', {
                'block_num_or_id': block_num_or_id
            }
        )

    async def get_block(self, block_num_or_id):
        return await self.post(
            '/chain/get_block', {
                'block_num_or_id': block_num_or_id
            }
        )

    async def get_code(self, account_name: str):
        return await self.post(
            '/chain/get_code', {
                'account_name': account_name
            }
        )

    async def get_currency_balance(self, code: str, account: str, symbol: str):
        return await self.post(
            '/chain/get_currency_balance', {
                'code': code,
                'account': account,
                'symbol': symbol
            }
        )

    async def get_currency_stats(self, code: str, symbol: str):
        return await self.post(
            '/chain/get_currency_stats', {
                'code': code,
                'symbol': symbol
            }
        )

    async def get_info(self):
        return await self.post('/chain/get_info')

    async def get_chain_id(self):
        if not self._chain_id:
            info = await self.get_info()
            self._chain_id = binascii.unhexlify(info['chain_id'])
        return self._chain_id

    async def get_head_block(self):
        info = await self.get_info()
        return await self.get_block(info['head_block_num'])

    async def get_producer_schedule(self):
        return await self.post('/chain/get_producer_schedule')

    async def get_producers(self, json=True, lower_bound='', limit=50):
        return await self.post(
            '/chain/get_producers', {
                'json': json,
                'lower_bound': lower_bound,
                'limit': limit
            }
        )

    async def get_raw_code_and_abi(self, account_name: str):
        return await self.post(
            '/chain/get_raw_code_and_abi', {
                'account_name': account_name
            }
        )

    async def get_raw_abi(self, account_name: str):
        response = await self.post(
            '/chain/get_raw_code_and_abi', {
                'account_name': account_name
            }
        )
        return {
            'account_name': response.get('account_name'),
            'abi': base64.b64decode(response.get('abi'))
        }

    async def get_table_rows(
        self, code, scope, table, table_key='', lower_bound='', upper_bound='',
        index_position=1, key_type='', limit=10, reverse=False,
        show_payer=False, json=True
    ):
        return await self.post(
            '/chain/get_table_rows', {
                'json': json,
                'code': code,
                'scope': scope,
                'table': table,
                'table_key': table_key,
                'lower_bound': lower_bound,
                'upper_bound': upper_bound,
                'index_position': index_position,
                'key_type': key_type,
                'limit': limit,
                'reverse': reverse,
                'show_payer': show_payer
            }
        )

    async def get_table_by_scope(
        self, code, table, lower_bound='', upper_bound='', limit=10
    ):
        return await self.post(
            '/chain/get_table_by_scope', {
                'code': code,
                'table': table,
                'lower_bound': lower_bound,
                'upper_bound': upper_bound,
                'limit': limit
            }
        )

    async def get_required_keys(self, transaction, available_keys):
        return await self.post(
            '/chain/get_required_keys', {
                'transaction': transaction,
                'available_keys': available_keys
            }
        )

    async def sign_and_push_transaction(
        self,
        transaction: EosTransaction,
        *,
        context_free_bytes: bytes = bytes(32),
        keys: List[EosKey] = [],
        use_stored = True
    ):
        """ Specify use_stored=False to refetch the ABI if that contract is already stored.""" 
        for action in transaction.actions:
            if isinstance(action.data, dict):
                abi_bin = await self.abi_json_to_bin(
                    action, use_stored
                )
                action.data = binascii.unhexlify(abi_bin['binargs'])

        chain_id = await self.get_chain_id()
        serialized_transaction = serializer.serialize(transaction)

        digest = hashlib.sha256(
            b''.join((chain_id, serialized_transaction, context_free_bytes))
        ).digest()

        return await self.push_transaction(
            signatures=[key.sign(digest) for key in keys],
            serialized_transaction=(
                binascii.hexlify(serialized_transaction).decode()
            )
        )


    async def smart_sign_and_push_transaction(self, transaction: EosTransaction, *, context_free_bytes: bytes = bytes(32), keys: List[EosKey] = []
    ):
        """ Signs and pushes a transaction, refetching the ABI if needed on fail.""" 
        try:
            await self.sign_and_push_transaction(transaction, context_free_bytes, keys, True)
        except Exception as e:
            await self.sign_and_push_transaction(transaction, context_free_bytes, keys, False)

    async def push_actions(self, actions: List[EosAction], api_endpoints: List[str], account: EosAccount, retry = 5):
        """Function to try and push an action across a list of endpoints"""
        if len(api_endpoints) == 0:
            raise Exception("Please pass at least one API endpoint")
        index = (secrets.randbelow(len(api_endpoints)))
        for i in range(0, retry):
            try:
                endpoint = api_endpoints[index % len(api_endpoints)]
                self.URL = endpoint
                ref_block = await self.get_head_block()
                transaction = EosTransaction(
                    ref_block_num=ref_block["block_num"] & 65535,
                    ref_block_prefix=ref_block["ref_block_prefix"],
                    actions=actions)
                await self.sign_and_push_transaction(transaction, keys=[account.key], use_stored={i < 2})
                return True
            except Exception as e:
                try:
                    if e.args[0]['code'] == 3050003:
                        return False
                except:
                    #print("Failed, will try again; ", i, e)
                    index += 1
                    await asyncio.sleep(0.19 * i)

    async def push_transaction(self, signatures, serialized_transaction):
        return await self.post(
            '/chain/push_transaction', {
                'signatures': signatures,
                'compression': 0,
                'packed_context_free_data': '',
                'packed_trx': serialized_transaction
            }
        )

    async def get_db_size(self):
        return await self.post('/db_size/get')

    async def get_actions(self, account_name: str, pos=None, offset=None):
        return await self.post(
            '/history/get_actions', {
                'account_name': account_name,
                'pos': pos,
                'offset': offset
            }
        )

    async def get_transaction(self, tx_id: str, block_num_hint=None):
        return await self.post(
            '/history/get_transaction', {
                'id': tx_id,
                'block_num_hint': block_num_hint
            }
        )

    async def get_key_accounts(self, public_key):
        return await self.post(
            '/history/get_key_accounts', {
                'public_key': public_key
            }
        )

    async def get_controlled_accounts(self, account_name: str):
        return await self.post(
            '/history/get_controlled_accounts', {
                'controlling_account': account_name
            }
        )



from .abi import (
    UInt8, UInt16, UInt32, UInt64, Int8, Int16, Int32, Int64, VarUInt, Float32,
    Float64, TimePointSec, TimePoint, Name, AbiBytes, BaseAbiObject,
    is_abi_object
)  # noqa

from .authority import (
    EosPermissionLevel, EosKeyWeight, EosPermissionLevelWeight, EosWaitWeight,
    EosAuthority
)  # noqa

from .transaction import (
    AbiActionPayload, EosAction, EosTransaction
)  # noqa


__all__ = [
    # base ABI types
    'UInt8', 'UInt16', 'UInt32', 'UInt64', 'Int8', 'Int16', 'Int32', 'Int64',
    'VarUInt', 'Float32', 'Float64', 'TimePointSec', 'TimePoint', 'Name',
    'AbiBytes', 'BaseAbiObject', 'is_abi_object',

    # authority
    'EosPermissionLevel', 'EosKeyWeight', 'EosPermissionLevelWeight',
    'EosWaitWeight', 'EosAuthority',

    # transaction
    'AbiActionPayload', 'EosAction', 'EosTransaction'
]



from abc import ABC, abstractmethod
from datetime import datetime, timezone
import inspect
import struct
from typing import Any, List, Tuple, Type, Union

from aioeosabi import types, exceptions


class BaseSerializer(ABC):
    @abstractmethod
    def serialize(self, value: Any) -> bytes:
        """Returns byte-encoded value"""
        pass  # pragma: no cover

    @abstractmethod
    def deserialize(self, value: bytes) -> Tuple[int, Any]:
        """
        Returns a tuple containing length of original data and deserialized
        value
        """
        pass  # pragma: no cover


class BasicTypeSerializer(BaseSerializer):
    """
    Serializes basic types such as integers and floats using ``struct`` module

    :params fmt: format string, please refer to documentation for `struct
                 module <https://docs.python.org/3/library/struct.html>`_
    """

    def __init__(self, fmt: Union[bytes, str] = ''):
        assert fmt, 'provide valid fmt value'
        self.fmt = fmt

    def serialize(self, value: Any) -> bytes:
        return struct.pack(self.fmt, value)

    def deserialize(self, value: bytes) -> Tuple[int, Any]:
        length = struct.calcsize(self.fmt)
        values = struct.unpack_from(self.fmt, value)
        return length, values[0]


class AbiNameSerializer(BasicTypeSerializer):
    """
    Serializer for ABI names. ABI names can only contain these characters:
    ``.12345abcdefghijklmnopqrstuvwxyz``. Maximum length is 13 chars.
    """

    def __init__(self):
        self.alphabet = '.12345abcdefghijklmnopqrstuvwxyz'
        self.fmt = 'Q'

    def serialize(self, value: str) -> bytes:
        # Run preliminary checks
        if len(value) > 13:
            raise exceptions.EosSerializerAbiNameTooLongException

        if any(x not in self.alphabet for x in value):
            raise exceptions.EosSerializerAbiNameInvalidCharactersException

        name = sum(
            (self.alphabet.index(char) & 0x1F) << (64-5 * (index + 1))
            for index, char in enumerate(value[:12])
        )
        if len(value) > 12:
            name |= self.alphabet.index(value[12]) & 0x0F
        return super().serialize(name)

    def deserialize(self, value: bytes) -> Tuple[int, str]:
        length, decoded_value = super().deserialize(value)
        name = ['.'] * 13
        i = 12
        while i >= 0:
            # 13th character is encoded on last 2 bytes
            mask = 0x0F if i == 12 else 0x1F
            name[i] = self.alphabet[decoded_value & mask]
            decoded_value >>= 4 if i == 12 else 5
            i -= 1
        return length, ''.join(name).rstrip('.')


class VarUIntSerializer(BaseSerializer):
    """
    Serializer for ABI VarUInt type. This type has different length based on
    how many bytes are required to encode given integer.
    """

    def serialize(self, value: int) -> bytes:
        # TODO: copied over from libeospy, I'm 99% sure this is overcomplicated
        array = bytearray()
        val = int(value)
        buf = int((val) & 0x7f)
        val >>= 7
        buf |= (val > 0) << 7
        array.append(int(buf))
        while val:
            buf = int((val) & 0x7f)
            val >>= 7
            buf |= (val > 0) << 7
            array.append(int(buf))
        return bytes(array)

    def deserialize(self, value: bytes) -> Tuple[int, int]:
        # TODO: copied over from libeospy, I'm 99% sure this is overcomplicated
        shift = 0
        result = 0
        cursor = 0
        while True:
            tmp = value[cursor]
            result |= (tmp & 0x7f) << shift
            shift += 7
            cursor += 1
            if not(tmp & 0x80):
                break
        return cursor, result


class AbiBytesSerializer(BaseSerializer):
    """
    Serializer for ABI bytes type. Serialized value consists of raw bytes
    prefixed with payload size encoded as VarUInt.
    """

    def serialize(self, value: bytes) -> bytes:
        assert isinstance(value, bytes), 'Provide binary format'
        prefix = VarUIntSerializer().serialize(len(value))
        return b''.join((prefix, value))

    def deserialize(self, value: bytes) -> Tuple[int, bytes]:
        prefix_length, prefix = VarUIntSerializer().deserialize(value)
        overall_length = prefix_length + prefix
        return overall_length, value[prefix_length:overall_length]


class AbiTimePointSerializer(BasicTypeSerializer):
    """
    Serializer for ABI TimePoint type. Encodes timestamp with milisecond
    precision.
    """

    def __init__(self):
        self.fmt = 'Q'

    def serialize(self, value: datetime) -> bytes:
        return super().serialize(int(value.timestamp() * 1000))

    def deserialize(self, value: bytes) -> Tuple[int, datetime]:
        length, decoded_value = super().deserialize(value)
        return (
            length, datetime.fromtimestamp(decoded_value / 1000, timezone.utc)
        )


class AbiTimePointSecSerializer(BasicTypeSerializer):
    """
    Serializer for ABI TimePointSec type. It's essentially a timestamp.
    """

    def __init__(self):
        self.fmt = 'I'

    def serialize(self, value: datetime) -> bytes:
        return super().serialize(int(value.timestamp()))

    def deserialize(self, value: bytes) -> Tuple[int, datetime]:
        length, decoded_value = super().deserialize(value)
        return length, datetime.fromtimestamp(decoded_value, timezone.utc)


class AbiStringSerializer(BaseSerializer):
    """
    Serializer for ABI String type. String format is similar to bytes as it's
    prefixed with length but it's comprised of ASCII codes for each character
    packed in binary format.
    """
    def serialize(self, value: str) -> bytes:
        return AbiBytesSerializer().serialize(value.encode())

    def deserialize(self, value: bytes) -> Tuple[int, str]:
        length, value = AbiBytesSerializer().deserialize(value)
        return length, value.decode()


class AbiObjectSerializer(BaseSerializer):
    def __init__(self, abi_class: Type):
        self.abi_class = abi_class

    def serialize(self, value: types.BaseAbiObject) -> bytes:
        assert issubclass(value.__class__, self.abi_class)
        return b''.join(
            serialize(
                getattr(value, field),
                value.__dataclass_fields__[field].type,
            )
            for field in value._serializable_fields()
        )

    def deserialize(self, value: bytes) -> Tuple[int, types.BaseAbiObject]:
        cursor = 0
        values = {}
        for field in self.abi_class._serializable_fields():
            length, values[field] = deserialize(
                value[cursor:],
                self.abi_class.__dataclass_fields__[field].type
            )
            cursor += length
        return cursor, self.abi_class(**values)


class AbiActionPayloadSerializer(BaseSerializer):
    def serialize(self, value: types.AbiActionPayload) -> bytes:
        assert not isinstance(value, dict), 'Convert data to ABI format first'
        if types.is_abi_object(type(value)):
            # mypy won't recognize that as a type check apparently
            serializer = AbiObjectSerializer(type(value))
            value = serializer.serialize(value)  # type: ignore
        assert isinstance(value, bytes)
        return AbiBytesSerializer().serialize(value)

    def deserialize(self, value: bytes) -> Tuple[int, bytes]:
        # TODO: figure out how to convert it back to BaseAbiObject or dict
        return AbiBytesSerializer().deserialize(value)


TYPE_MAPPING = {
    types.UInt8: BasicTypeSerializer('B'),
    types.UInt16: BasicTypeSerializer('H'),
    types.UInt32: BasicTypeSerializer('I'),
    types.UInt64: BasicTypeSerializer('Q'),
    types.Int8: BasicTypeSerializer('b'),
    types.Int16: BasicTypeSerializer('h'),
    types.Int32: BasicTypeSerializer('i'),
    types.Int64: BasicTypeSerializer('q'),
    types.Float32: BasicTypeSerializer('f'),
    types.Float64: BasicTypeSerializer('d'),
    types.Name: AbiNameSerializer(),
    types.VarUInt: VarUIntSerializer(),
    types.AbiBytes: AbiBytesSerializer(),
    types.AbiActionPayload: AbiActionPayloadSerializer(),  # type: ignore
    types.TimePoint: AbiTimePointSerializer(),
    types.TimePointSec: AbiTimePointSecSerializer(),
    str: AbiStringSerializer()
}


class AbiListSerializer(BaseSerializer):
    """
    Serializer for ABI List type. In binary format, it basically looks like
    this: ``[count][item 1][item 2]...``
    """

    def __init__(self, list_type: Type):
        self.eos_type = list_type.__args__[0]
        self.item_serializer = get_abi_type_serializer(self.eos_type)

    def serialize(self, value: List[Any]) -> bytes:
        return b''.join([
            VarUIntSerializer().serialize(len(value)),
            *(self.item_serializer.serialize(x) for x in value)
        ])

    def deserialize(self, value: bytes) -> Tuple[int, List[Any]]:
        # List always starts with a VarUInt representing item count
        prefix_length, count = VarUIntSerializer().deserialize(value)

        abi_bytes = value[prefix_length:]
        overall_length = prefix_length
        values = []

        for _ in range(count):
            # try to decode bytes as expected EOS type
            length, decoded_value = self.item_serializer.deserialize(abi_bytes)

            # advance buffer by read and decoded bytes
            overall_length += length
            abi_bytes = abi_bytes[length:]

            # append decoded value to values list
            values.append(decoded_value)
        return overall_length, values


def get_abi_type_serializer(abi_type: Type) -> BaseSerializer:
    if abi_type in TYPE_MAPPING:
        return TYPE_MAPPING[abi_type]
    elif getattr(abi_type, '_name', None) == 'List':
        return AbiListSerializer(abi_type)
    elif types.is_abi_object(abi_type):
        return AbiObjectSerializer(abi_type)

    # if type is not supported, raise an Exception
    raise exceptions.EosSerializerUnsupportedTypeException(abi_type)


def serialize(value: Any, abi_type: Type = None) -> bytes:
    """Serializes ABI values to binary format"""
    if not abi_type:
        is_class = inspect.isclass(value)
        abi_type = value.__class__ if is_class else type(value)
    return get_abi_type_serializer(abi_type).serialize(value)


def deserialize(value: bytes, abi_class: Type) -> Tuple[int, Any]:
    """Deserializes ABI values from binary format"""
    return get_abi_type_serializer(abi_class).deserialize(value)
